from typing import List
from unittest import mock
from unittest.mock import AsyncMock

from promptl_ai import (
    Adapter,
    AssistantMessage,
    Error,
    ErrorPosition,
    MessageLike,
    PromptlError,
    SystemMessage,
    TextContent,
    UserMessage,
    openai,
)

from latitude_sdk import RenderChainOptions, RenderChainResult
from tests.utils import TestCase, fixtures


class TestRenderChain(TestCase):
    async def test_success_without_adapter(self):
        on_step_mock = AsyncMock(
            side_effect=[
                "The definitive answer is 42.",
                "Goodbye!",
            ]
        )
        prompt = fixtures.PROMPT
        options = RenderChainOptions(
            parameters={"question": "What is PromptL"},
            adapter=Adapter.Default,
        )

        result = await self.sdk.prompts.render_chain(prompt, on_step_mock, options)

        expected_messages: List[MessageLike] = [
            SystemMessage(content=[TextContent(text="You are a helpful assistant.")]),
            UserMessage(content=[TextContent(text="What is PromptL")]),
            AssistantMessage(content=[TextContent(text="The definitive answer is 42.")]),
            SystemMessage(content=[TextContent(text="Now say bye.")]),
            AssistantMessage(content=[TextContent(text="Goodbye!")]),
        ]
        expected_config = {
            "provider": "OpenAI",
            "model": "gpt-4o-mini",
            "temperature": 0.5,
            "maxTokens": 1024,
            "topP": 0.9,
        }

        self.assertEqual(
            result,
            RenderChainResult(
                messages=expected_messages,
                config=expected_config,
            ),
        )
        self.assertEqual(
            on_step_mock.call_args_list[0][0],
            (
                expected_messages[:2],
                expected_config,
            ),
        )
        self.assertEqual(
            on_step_mock.call_args_list[1][0],
            (
                expected_messages[:-1],
                expected_config,
            ),
        )
        self.assertEqual(on_step_mock.await_count, 2)

    async def test_success_with_adapter(self):
        on_step_mock = AsyncMock(
            side_effect=[
                "The definitive answer is 42.",
                "Goodbye!",
            ]
        )
        prompt = fixtures.PROMPT
        options = RenderChainOptions(
            parameters={"question": "What is PromptL"},
            adapter=Adapter.OpenAI,
        )

        result = await self.sdk.prompts.render_chain(prompt, on_step_mock, options)

        expected_messages: List[MessageLike] = [
            openai.SystemMessage(content="You are a helpful assistant."),
            openai.UserMessage(content=[openai.TextContent(text="What is PromptL")]),
            openai.AssistantMessage(content=[openai.TextContent(text="The definitive answer is 42.")]),
            openai.SystemMessage(content="Now say bye."),
            openai.AssistantMessage(content=[openai.TextContent(text="Goodbye!")]),
        ]
        expected_config = {
            "provider": "OpenAI",
            "model": "gpt-4o-mini",
            "temperature": 0.5,
            "max_tokens": 1024,
            "top_p": 0.9,
        }

        self.assertEqual(
            result,
            RenderChainResult(
                messages=expected_messages,
                config=expected_config,
            ),
        )
        self.assertEqual(
            on_step_mock.call_args_list[0][0],
            (
                expected_messages[:2],
                expected_config,
            ),
        )
        self.assertEqual(
            on_step_mock.call_args_list[1][0],
            (
                expected_messages[:-1],
                expected_config,
            ),
        )
        self.assertEqual(on_step_mock.await_count, 2)

    async def test_fails(self):
        on_step_mock = AsyncMock(
            side_effect=[
                "The definitive answer is 42.",
                "Goodbye!",
            ]
        )
        prompt = fixtures.PROMPT
        options = RenderChainOptions(
            adapter=Adapter.Default,
        )

        with self.assertRaises(PromptlError) as context:
            await self.sdk.prompts.render_chain(prompt, on_step_mock, options)

        self.assertEqual(
            context.exception,
            PromptlError(
                Error.model_construct(
                    name="CompileError",
                    code="variable-not-declared",
                    message="Variable 'question' is not declared",
                    start=ErrorPosition(line=11, column=14, character=141),
                    end=ErrorPosition(line=11, column=22, character=149),
                    frame=mock.ANY,
                )
            ),
        )



import json
from typing import List, cast

import httpx

from latitude_sdk import CreateLogOptions, CreateLogResult
from tests.utils import TestCase, fixtures


class TestCreateLog(TestCase):
    async def test_success_global_options(self):
        path = "prompt-path"
        messages = self.create_conversation(4)
        options = CreateLogOptions(response="response")
        endpoint = f"/projects/{self.project_id}/versions/{self.version_uuid}/documents/logs"
        endpoint_mock = self.gateway_mock.post(endpoint).mock(
            return_value=httpx.Response(200, json=fixtures.LOG_RESPONSE)
        )

        result = await self.sdk.logs.create(path, messages, options)
        request, _ = endpoint_mock.calls.last

        self.assert_requested(
            request,
            method="POST",
            endpoint=endpoint,
            body={
                "path": path,
                "messages": [json.loads(message.model_dump_json()) for message in messages],
                "response": options.response,
            },
        )
        self.assertEqual(endpoint_mock.call_count, 1)
        self.assertEqual(result, CreateLogResult(**dict(fixtures.LOG)))

    async def test_success_overrides_options(self):
        path = "prompt-path"
        messages = self.create_conversation(4)
        options = CreateLogOptions(project_id=21, version_uuid="version-uuid", response="response")
        endpoint = f"/projects/{options.project_id}/versions/{options.version_uuid}/documents/logs"
        endpoint_mock = self.gateway_mock.post(endpoint).mock(
            return_value=httpx.Response(200, json=fixtures.LOG_RESPONSE)
        )

        result = await self.sdk.logs.create(path, messages, options)
        request, _ = endpoint_mock.calls.last

        self.assert_requested(
            request,
            method="POST",
            endpoint=endpoint,
            body={
                "path": path,
                "messages": [json.loads(message.model_dump_json()) for message in messages],
                "response": options.response,
            },
        )
        self.assertEqual(endpoint_mock.call_count, 1)
        self.assertEqual(result, CreateLogResult(**dict(fixtures.LOG)))

    async def test_success_default_version_uuid(self):
        self.sdk._options.version_uuid = None  # pyright: ignore [reportPrivateUsage]
        path = "prompt-path"
        messages = self.create_conversation(4)
        options = CreateLogOptions(response="response")
        endpoint = f"/projects/{self.project_id}/versions/live/documents/logs"
        endpoint_mock = self.gateway_mock.post(endpoint).mock(
            return_value=httpx.Response(200, json=fixtures.LOG_RESPONSE)
        )

        result = await self.sdk.logs.create(path, messages, options)
        request, _ = endpoint_mock.calls.last

        self.assert_requested(
            request,
            method="POST",
            endpoint=endpoint,
            body={
                "path": path,
                "messages": [json.loads(message.model_dump_json()) for message in messages],
                "response": options.response,
            },
        )
        self.assertEqual(endpoint_mock.call_count, 1)
        self.assertEqual(result, CreateLogResult(**dict(fixtures.LOG)))

    async def test_fails(self):
        path = "prompt-path"
        messages = self.create_conversation(4)
        options = CreateLogOptions(response="response")
        endpoint = f"/projects/{self.project_id}/versions/{self.version_uuid}/documents/logs"
        endpoint_mock = self.gateway_mock.post(endpoint).mock(
            return_value=httpx.Response(500, json=fixtures.ERROR_RESPONSE)
        )

        with self.assertRaisesRegex(type(fixtures.ERROR), fixtures.ERROR.message):
            await self.sdk.logs.create(path, messages, options)
        requests = cast(List[httpx.Request], [request for request, _ in endpoint_mock.calls])  # type: ignore

        [
            self.assert_requested(
                request,
                method="POST",
                endpoint=endpoint,
                body={
                    "path": path,
                    "messages": [json.loads(message.model_dump_json()) for message in messages],
                    "response": options.response,
                },
            )
            for request in requests
        ]
        self.assertEqual(endpoint_mock.call_count, self.internal_options.retries)



import sys

from sh import pyright, ruff  # type: ignore

files = sys.argv[1:] or ["."]

pyright(*files, _out=sys.stdout)
ruff("check", *files, _out=sys.stdout)
ruff("format", "--check", *files, _out=sys.stdout)



from typing import Optional, Union

from latitude_sdk.util import Model, StrEnum


# NOTE: Incomplete list
class ApiErrorCodes(StrEnum):
    # LatitudeErrorCodes
    NotFoundError = "NotFoundError"

    # RunErrorCodes
    AIRunError = "ai_run_error"

    # ApiErrorCodes
    HTTPException = "http_exception"
    InternalServerError = "internal_server_error"


UNEXPECTED_ERROR_CODES = [
    str(ApiErrorCodes.HTTPException),
    str(ApiErrorCodes.InternalServerError),
]


class ApiErrorDbRef(Model):
    entity_uuid: str
    entity_type: str


class ApiError(Exception):
    status: int
    code: str  # NOTE: Cannot be ApiErrorCodes because the list is incomplete
    message: str
    response: str
    db_ref: Optional[ApiErrorDbRef]

    def __init__(
        self,
        status: int,
        code: Union[str, ApiErrorCodes],
        message: str,
        response: str,
        db_ref: Optional[ApiErrorDbRef] = None,
    ):
        message = self._exception_message(status, str(code), message)
        super().__init__(message)

        self.status = status
        self.code = str(code)
        self.message = message
        self.response = response
        self.db_ref = db_ref

    @staticmethod
    def _exception_message(status: int, code: str, message: str) -> str:
        if code in UNEXPECTED_ERROR_CODES:
            return f"Unexpected API Error: {status} {message}"

        return message

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, ApiError)
            and self.status == other.status
            and self.code == other.code
            and self.message == other.message
            and self.response == other.response
            and self.db_ref == other.db_ref
        )



from unittest import mock

from promptl_ai import (
    Adapter,
    Error,
    ErrorPosition,
    PromptlError,
    SystemMessage,
    TextContent,
    UserMessage,
    anthropic,
    openai,
)

from latitude_sdk import RenderPromptOptions, RenderPromptResult
from tests.utils import TestCase, fixtures


class TestRenderPrompt(TestCase):
    async def test_success_without_adapter(self):
        prompt = fixtures.PROMPT.content
        options = RenderPromptOptions(
            parameters={"question": "What is PromptL"},
            adapter=Adapter.Default,
        )

        result = await self.sdk.prompts.render(prompt, options)

        self.assertEqual(
            result,
            RenderPromptResult(
                messages=[
                    SystemMessage(content=[TextContent(text="You are a helpful assistant.")]),
                    UserMessage(content=[TextContent(text="What is PromptL")]),
                ],
                config={
                    "provider": "OpenAI",
                    "model": "gpt-4o-mini",
                    "temperature": 0.5,
                    "maxTokens": 1024,
                    "topP": 0.9,
                },
            ),
        )

    async def test_success_with_adapter_singlepart(self):
        prompt = fixtures.PROMPT.content
        options = RenderPromptOptions(
            parameters={"question": "What is PromptL"},
            adapter=Adapter.OpenAI,
        )

        result = await self.sdk.prompts.render(prompt, options)

        self.assertEqual(
            result,
            RenderPromptResult(
                messages=[
                    openai.SystemMessage(content="You are a helpful assistant."),
                    openai.UserMessage(content=[openai.TextContent(text="What is PromptL")]),
                ],
                config={
                    "provider": "OpenAI",
                    "model": "gpt-4o-mini",
                    "temperature": 0.5,
                    "max_tokens": 1024,
                    "top_p": 0.9,
                },
            ),
        )

    async def test_success_with_adapter_multipart(self):
        parts = fixtures.PROMPT.content.split("</step>")
        prompt = "</step>".join(
            [
                parts[0]
                + """
<user>
    This is text.
    <content-image>This is an image.</content-image>
    <content-file mime="application/pdf">This is a PDF.</content-file>
</user>
""",
                *parts[1:],
            ]
        )
        options = RenderPromptOptions(
            parameters={"question": "What is PromptL"},
            adapter=Adapter.Anthropic,
        )

        result = await self.sdk.prompts.render(prompt, options)

        self.assertEqual(
            result,
            RenderPromptResult(
                messages=[
                    anthropic.UserMessage(content=[anthropic.TextContent(text="What is PromptL")]),
                    anthropic.UserMessage(
                        content=[
                            anthropic.TextContent(text="This is text."),
                            anthropic.ImageContent(
                                source=anthropic.ContentSource(
                                    type="base64", media_type="image/png", data="This is an image."
                                )
                            ),
                            anthropic.DocumentContent(
                                source=anthropic.ContentSource(
                                    type="base64", media_type="application/pdf", data="This is a PDF."
                                )
                            ),
                        ]
                    ),
                ],
                config={
                    "provider": "OpenAI",
                    "model": "gpt-4o-mini",
                    "temperature": 0.5,
                    "max_tokens": 1024,
                    "top_p": 0.9,
                    "system": [{"type": "text", "text": "You are a helpful assistant."}],
                },
            ),
        )

    async def test_fails(self):
        prompt = fixtures.PROMPT.content
        options = RenderPromptOptions(
            adapter=Adapter.Default,
        )

        with self.assertRaises(PromptlError) as context:
            await self.sdk.prompts.render(prompt, options)

        self.assertEqual(
            context.exception,
            PromptlError(
                Error.model_construct(
                    name="CompileError",
                    code="variable-not-declared",
                    message="Variable 'question' is not declared",
                    start=ErrorPosition(line=11, column=14, character=141),
                    end=ErrorPosition(line=11, column=22, character=149),
                    frame=mock.ANY,
                )
            ),
        )
